@@ -25,8 +25,6 @@
 #include "internal.h"
 
 void libcouchbase_purge_single_server(libcouchbase_server_t *server,
-                                      ringbuffer_t *stream,
-                                      ringbuffer_t *cookies,
                                       hrtime_t tmo,
                                       hrtime_t now,
                                       libcouchbase_error_t error)
@@ -38,15 +36,25 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
     libcouchbase_uint32_t packetsize;
     char *keyptr;
     libcouchbase_t root = server->instance;
-    int writing = (stream == &server->cmd_log);
     ringbuffer_t rest;
+    ringbuffer_t *stream = &server->cmd_log;
+    ringbuffer_t *cookies;
+    ringbuffer_t *mirror = NULL; /* mirror buffer should be purged with main stream */
     libcouchbase_size_t send_size = libcouchbase_ringbuffer_get_nbytes(&server->output);
     libcouchbase_size_t stream_size = libcouchbase_ringbuffer_get_nbytes(stream);
 
+    if (server->connected) {
+        cookies = &server->output_cookies;
+    } else {
+        cookies = &server->pending_cookies;
+        mirror = &server->pending;
+    }
+
     assert(libcouchbase_ringbuffer_initialize(&rest, 1024));
 
     do {
         int allocated = 0;
+        libcouchbase_uint32_t headersize;
         nr = libcouchbase_ringbuffer_peek(cookies, &ct, sizeof(ct));
         if (nr != sizeof(ct) || ct.start + tmo > now) {
             break;
@@ -55,20 +63,22 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
         if (nr != sizeof(req)) {
             break;
         }
+        packetsize = (libcouchbase_uint32_t)sizeof(req) + ntohl(req.request.bodylen);
+        if (libcouchbase_ringbuffer_get_nbytes(stream) < packetsize) {
+            break;
+        }
 
         libcouchbase_ringbuffer_consumed(cookies, sizeof(ct));
 
         assert(nr == sizeof(req));
         packet = stream->read_head;
-        packetsize = ntohl(req.request.bodylen) + (libcouchbase_uint32_t)sizeof(req);
 
         if (server->instance->histogram) {
             libcouchbase_record_metrics(server->instance, now - ct.start,
                                         req.request.opcode);
         }
 
-        if (writing && (stream_size > send_size) &&
-            ((stream_size - packetsize) < send_size)) {
+        if (server->connected && stream_size > send_size && (stream_size - packetsize) < send_size) {
             /* Copy the rest of the current packet into the
                temporary stream */
 
@@ -81,18 +91,18 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
             send_size -= nbytes;
         }
         stream_size -= packetsize;
-        /* @todo fixme.. I don't need the value to be in the continous part.. */
+        headersize = (libcouchbase_uint32_t)sizeof(req) + req.request.extlen + htons(req.request.keylen);
         if (!libcouchbase_ringbuffer_is_continous(stream,
                                                   RINGBUFFER_READ,
-                                                  packetsize)) {
-            packet = malloc(packetsize);
+                                                  headersize)) {
+            packet = malloc(headersize);
             if (packet == NULL) {
                 libcouchbase_error_handler(server->instance, LIBCOUCHBASE_ENOMEM, NULL);
                 abort();
             }
 
-            nr = libcouchbase_ringbuffer_peek(stream, packet, packetsize);
-            if (nr != packetsize) {
+            nr = libcouchbase_ringbuffer_peek(stream, packet, headersize);
+            if (nr != headersize) {
                 libcouchbase_error_handler(server->instance, LIBCOUCHBASE_EINTERNAL,
                                            NULL);
                 free(packet);
@@ -170,11 +180,11 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
             root->callbacks.remove(root, ct.cookie,
                                    error,
                                    keyptr, ntohs(req.request.keylen));
-             break;
+            break;
 
         case PROTOCOL_BINARY_CMD_INCREMENT:
         case PROTOCOL_BINARY_CMD_DECREMENT:
-            root->callbacks.arithmetic(root,ct.cookie,
+            root->callbacks.arithmetic(root, ct.cookie,
                                        error,
                                        keyptr, ntohs(req.request.keylen), 0, 0);
             break;
@@ -208,6 +218,18 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
                                      error, NULL, 0, NULL, 0);
             }
             break;
+
+        case PROTOCOL_BINARY_CMD_VERSION:
+            root->callbacks.version(root, ct.cookie, server->authority,
+                                    error, NULL, 0);
+            if (libcouchbase_lookup_server_with_command(root,
+                                                        PROTOCOL_BINARY_CMD_VERSION,
+                                                        req.request.opaque,
+                                                        server) < 0) {
+                root->callbacks.version(root, ct.cookie, NULL, error, NULL, 0);
+            }
+            break;
+
         default:
             abort();
         }
@@ -217,9 +239,12 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
         }
 
         libcouchbase_ringbuffer_consumed(stream, packetsize);
+        if (mirror) {
+            libcouchbase_ringbuffer_consumed(mirror, packetsize);
+        }
     } while (1); /* CONSTCOND */
 
-    if (writing) {
+    if (server->connected) {
         /* Preserve the rest of the stream */
         libcouchbase_size_t nbytes = libcouchbase_ringbuffer_get_nbytes(stream);
         send_size = libcouchbase_ringbuffer_get_nbytes(&server->output);
@@ -238,20 +263,14 @@ void libcouchbase_purge_single_server(libcouchbase_server_t *server,
     if (libcouchbase_ringbuffer_peek(cookies, &ct, sizeof(ct)) == sizeof(ct)) {
         server->next_timeout = ct.start;
     }
+
+    libcouchbase_maybe_breakout(server->instance);
 }
 
 libcouchbase_error_t libcouchbase_failout_server(libcouchbase_server_t *server,
                                                  libcouchbase_error_t error)
 {
-    if (server->connected) {
-        libcouchbase_purge_single_server(server, &server->cmd_log,
-                                         &server->output_cookies,
-                                         0, gethrtime() + 1, error);
-    } else {
-        libcouchbase_purge_single_server(server, &server->pending,
-                                         &server->pending_cookies,
-                                         0, gethrtime() + 1, error);
-    }
+    libcouchbase_purge_single_server(server, 0, gethrtime() + 1, error);
 
     libcouchbase_ringbuffer_reset(&server->output);
     libcouchbase_ringbuffer_reset(&server->input);
@@ -335,10 +354,9 @@ static int get_local_address(libcouchbase_socket_t sock,
     socklen_t salen = sizeof(saddr);
 
     if ((getsockname(sock, (struct sockaddr *)&saddr, &salen) < 0) ||
-        (getnameinfo((struct sockaddr *)&saddr, salen, h, sizeof(h),
-                     p, sizeof(p), NI_NUMERICHOST | NI_NUMERICSERV) < 0) ||
-        (snprintf(buffer, bufsz, "%s;%s", h, p) < 0))
-    {
+            (getnameinfo((struct sockaddr *)&saddr, salen, h, sizeof(h),
+                         p, sizeof(p), NI_NUMERICHOST | NI_NUMERICSERV) < 0) ||
+            (snprintf(buffer, bufsz, "%s;%s", h, p) < 0)) {
         return 0;
     }
 
@@ -362,10 +380,9 @@ static int get_remote_address(libcouchbase_socket_t sock,
     socklen_t salen = sizeof(saddr);
 
     if ((getpeername(sock, (struct sockaddr *)&saddr, &salen) < 0) ||
-        (getnameinfo((struct sockaddr *)&saddr, salen, h, sizeof(h),
-                     p, sizeof(p), NI_NUMERICHOST | NI_NUMERICSERV) < 0) ||
-        (snprintf(buffer, bufsz, "%s;%s", h, p) < 0))
-    {
+            (getnameinfo((struct sockaddr *)&saddr, salen, h, sizeof(h),
+                         p, sizeof(p), NI_NUMERICHOST | NI_NUMERICSERV) < 0) ||
+            (snprintf(buffer, bufsz, "%s;%s", h, p) < 0)) {
         return 0;
     }
 
@@ -409,8 +426,8 @@ void libcouchbase_server_connected(libcouchbase_server_t *server)
         libcouchbase_ringbuffer_reset(&server->output_cookies);
         libcouchbase_ringbuffer_reset(&server->output);
         if (!libcouchbase_ringbuffer_append(&server->pending, &server->output) ||
-            !libcouchbase_ringbuffer_append(&server->pending_cookies, &server->output_cookies) ||
-            !libcouchbase_ringbuffer_append(&copy, &server->cmd_log)) {
+                !libcouchbase_ringbuffer_append(&server->pending_cookies, &server->output_cookies) ||
+                !libcouchbase_ringbuffer_append(&copy, &server->cmd_log)) {
             libcouchbase_error_handler(server->instance,
                                        LIBCOUCHBASE_ENOMEM,
                                        NULL);
@@ -468,24 +485,22 @@ static void server_connect_handler(libcouchbase_socket_t sock, short which, void
     server_connect(server);
 }
 
-static void server_connect(libcouchbase_server_t *server) {
+static void server_connect(libcouchbase_server_t *server)
+{
     int retry;
+    int save_errno;
+
     do {
         if (server->sock == INVALID_SOCKET) {
             /* Try to get a socket.. */
-            while (server->curr_ai != NULL) {
-                server->sock = server->instance->io->socket(server->instance->io,
-                                                            server->curr_ai->ai_family,
-                                                            server->curr_ai->ai_socktype,
-                                                            server->curr_ai->ai_protocol);
-                if (server->sock != INVALID_SOCKET) {
-                    break;
-                }
-                server->curr_ai = server->curr_ai->ai_next;
-            }
+            server->sock = libcouchbase_gai2sock(server->instance,
+                                                 &server->curr_ai,
+                                                 &save_errno);
         }
 
         if (server->curr_ai == NULL) {
+            /*TODO: Maybe check save_errno now? */
+
             /* this means we're not going to retry!! add an error here! */
             libcouchbase_failout_server(server, LIBCOUCHBASE_CONNECT_ERROR);
             return ;
@@ -500,15 +515,16 @@ static void server_connect(libcouchbase_server_t *server) {
             socket_connected(server);
             return ;
         } else {
-            switch (server->instance->io->error) {
-            case EINTR:
+            libcouchbase_connect_status_t connstatus =
+                libcouchbase_connect_status(server->instance->io->error);
+            switch (connstatus) {
+            case LIBCOUCHBASE_CONNECT_EINTR:
                 retry = 1;
                 break;
-            case EISCONN:
+            case LIBCOUCHBASE_CONNECT_EISCONN:
                 socket_connected(server);
                 return ;
-            case EWOULDBLOCK:
-            case EINPROGRESS: /* First call to connect */
+            case LIBCOUCHBASE_CONNECT_EINPROGRESS: /*first call to connect*/
                 server->instance->io->update_event(server->instance->io,
                                                    server->sock,
                                                    server->event,
@@ -516,24 +532,24 @@ static void server_connect(libcouchbase_server_t *server) {
                                                    server,
                                                    server_connect_handler);
                 return ;
-            case EALREADY: /* Subsequent calls to connect */
+            case LIBCOUCHBASE_CONNECT_EALREADY: /* Subsequent calls to connect */
                 return ;
 
-            default:
-                if (errno == ECONNREFUSED) {
+            case LIBCOUCHBASE_CONNECT_EFAIL:
+                if (server->curr_ai->ai_next) {
                     retry = 1;
                     server->curr_ai = server->curr_ai->ai_next;
-                } else {
-                    libcouchbase_failout_server(server,
-                                                LIBCOUCHBASE_CONNECT_ERROR);
-                    return ;
-                }
+                    server->instance->io->delete_event(server->instance->io,
+                                                       server->sock,
+                                                       server->event);
+                    server->instance->io->close(server->instance->io, server->sock);
+                    server->sock = INVALID_SOCKET;
+                    break;
+                } /* Else, we fallthrough */
 
-                server->instance->io->delete_event(server->instance->io,
-                                                   server->sock,
-                                                   server->event);
-                server->instance->io->close(server->instance->io, server->sock);
-                server->sock = INVALID_SOCKET;
+            default:
+                libcouchbase_failout_server(server, LIBCOUCHBASE_CONNECT_ERROR);
+                return;
             }
         }
     } while (retry);
@@ -573,7 +589,7 @@ void libcouchbase_server_initialize(libcouchbase_server_t *server, int servernum
     server->curr_ai = server->root_ai;
     server->sock = INVALID_SOCKET;
     if (error != 0) {
-       server->curr_ai = server->root_ai = NULL;
+        server->curr_ai = server->root_ai = NULL;
     }
 
     server->sasl_conn = NULL;
@@ -601,7 +617,7 @@ int libcouchbase_server_purge_implicit_responses(libcouchbase_server_t *c,
 {
     protocol_binary_request_header req;
     libcouchbase_size_t nr =  libcouchbase_ringbuffer_peek(&c->cmd_log, req.bytes,
-                                              sizeof(req));
+                                                           sizeof(req));
     /* There should at _LEAST_ be _ONE_ message in here! */
     assert(nr == sizeof(req));
     while (req.request.opaque < seqno) {
@@ -662,7 +678,7 @@ int libcouchbase_server_purge_implicit_responses(libcouchbase_server_t *c,
                                        errinfo);
             return -1;
 
-            }
+        }
         }
 
         libcouchbase_ringbuffer_consumed(&c->cmd_log, packetsize);
